------------------------------
--      Are you local?      --
------------------------------

local L = AceLibrary("AceLocale-2.2"):new("TableInspector")
local tablet = AceLibrary("Tablet-2.0")

local _G = _G
local type = _G.type
local tostring = _G.tostring
local loadstring = _G.loadstring
local pcall = _G.pcall
local tinsert = _G.tinsert
local sort = _G.sort
local format = _G.format
local IsControlKeyDown = _G.IsControlKeyDown
local argCheck = AceLibrary.argCheck

----------------------------
--      Localization      --
----------------------------

L:RegisterTranslations("enUS", function() return {
   ["Table Inspector"] = true,
   ["TableName[.SubTable[...]]"] = true,
	["%s is not a valid table."] = true,
   ["SHOW_STRING_COPY"] = true,
	["..Parent Table.."] = true,
   ["Strings"] = true,
   ["Numbers"] = true,
   ["Booleans"] = true,
   ["Tables"] = true,
   ["Functions"] = true,
   ["UserData"] = true,
   ["Other"] = true,
   ["Close"] = true,
   ["Click a table to view it. Ctrl-click to open in new window. Click other entry for access to copy."] = true,
} end)

---------------------------------
--      AceOptions Config      --
---------------------------------

local function CheckTable(v)
	local _,t = pcall(loadstring("return "..v))
	return type(t) == 'table'
end

local options = {
   name = L["Table Inspector"], type='text',
   desc = L["Table Inspector"],
   get = false,
   set = "NewTable",
   input = true,
   validate = CheckTable,
   usage = L["TableName[.SubTable[...]]"],
	error = L["%s is not a valid table."],
}

---------------------------------
--      Addon Declaration      --
---------------------------------
--~ TableInspector, ti
TableInspector = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDebug-2.0")
local TableInspector = TableInspector
TableInspector.revision = tonumber(string.sub("$Revision", 12, -3))
TableInspector.currentTable = nil -- used when a table is provided instead of a table's global name

------------------------------
--      Initialization      --
------------------------------

function TableInspector:OnInitialize()
   TableInspector:RegisterChatCommand({"/TableInspector", "/ti"}, options)
end

function TableInspector:OnEnable()
--~    code mostly from Prat: UrlCopy
   StaticPopupDialogs["SHOW_STRING_COPY"] = {  
      text = "%s",
      button1 = OKAY,
      hasEditBox = 1,
      hasWideEditBox = 1,
      showAlert = 1, 
      timeout = 0,
      whileDead = 1,

      OnShow = function()
         local editBox = _G[this:GetName().."WideEditBox"]
         editBox:SetText(_G[this:GetName().."Text"]:GetText()) 
         editBox:SetFocus()
         editBox:HighlightText(0)

         local button = _G[this:GetName().."Button1"]
         button:ClearAllPoints()
         button:SetPoint("CENTER", editBox, "CENTER", 0, -30)

         _G[this:GetName().."AlertIcon"]:Hide()
      end,
      EditBoxOnEscapePressed = function() this:GetParent():Hide() end,
      EditBoxOnEnterPressed = function() this:GetParent():Hide() end,
   }
end

function TableInspector:OnDisable()
   StaticPopupDialogs["SHOW_STRING_COPY"] = nil
end

------------------------------
--      Core Functions      --
------------------------------

--ugly hack unceremoniously ripped from Nargiddley, Kemayo should be shot for this - so he delcares
local function ScheduleClose(name)
   TableInspector:ScheduleEvent(function() 
      tablet:Close(name)
      tablet:Unregister(name) 
   end, 0.01)
end

local function spairs(t)
	local a = {}
	for n in pairs(t) do tinsert(a, n) end
	sort(a, function(l,r) return tostring(l):lower() < tostring(r):lower() end)
	local i = 0
	return function ()   -- iterator function
		i = i + 1
		return a[i], a[i] and t[a[i]]
	end
end

local function NextTable(self, name, oldName)
	if not IsControlKeyDown() then
		ScheduleClose(oldName)
	end
	self:NewTable(name)
end

-- click to open table
-- click super to return to previous table
-- ctrl+click to open table in new window

function TableInspector:NewTable(name)
	argCheck(self, name, 2, 'string', 'table')
	local t
	if type(name) == 'table' then
		TableInspector.currentTable = name
		t = name
		name = "TableInspector.currentTable"
	else
		t = loadstring("return "..name)()
	end
	assert(type(t) == 'table', "Bad argument #2 to TableInspector.NewTable (table or table name expected)")
	local tstr = tostring(t)
   local parent = name:match("^(.*)[%.%[]\"?.-\"?%]?$")
	
   if tablet:IsRegistered(tstr) then return end
   tablet:Register(tstr,
      'data', {},
		'minWidth', 300,
      'cantAttach', true,
      'clickable', true,
      'parent', "UIParent",
      'showTitleWhenDetached', true,
      'showHintWhenDetached', true,
      'dontHook', true,
      'strata', "DIALOG",
      'children', function()
         tablet:SetTitle(name)
         local cat = tablet:AddCategory('text', tstr, 'justify', "CENTER", "showWithoutChildren", true, 'wrap', true,
				'child_child_indentation', 20, 'child_textR', .93, 'child_textG', .77, 'child_textB', 0
			)
			if parent then
				cat:AddCategory('text', L["..Parent Table.."], 'justify', "CENTER", 'showWithoutChildren', true,
					'func', NextTable, 'arg1', self, 'arg2', parent, 'arg3', tstr
				)
			end
			local catTable = cat:AddCategory('text', L["Tables"], 'child_wrap', true)
         local cats = {
				['string'] = 	cat:AddCategory('text', L["Strings"], 'child_wrap', true),
				['number'] = 	cat:AddCategory('text', L["Numbers"]),
				['boolean'] = 	cat:AddCategory('text', L["Booleans"]),
				['function'] = cat:AddCategory('text', L["Functions"]),
				['userdata'] = cat:AddCategory('text', L["UserData"]),
			}
			local catOther = cat:AddCategory('text', L["Other"], 'child_wrap', true)
			for k,v in spairs(t) do
				local type_v = type(v)
				local type_k = type(k)
				local msg
				if type_k == 'number' then
					msg = "|cffffa0ff%d|r = "
				else
					msg = "|cffffa0ff%q|r = "
				end
				if type_v == 'string' or type_v == 'table' then
					msg = msg.."%q"
				else
					msg = msg.."%s"
				end
				msg = msg:format(tostring(k), tostring(v)):gsub("(%a+:)", "|cffd0ffd0%1|r")
				if type_v == 'table' then
					local newname
					local func = NextTable
					if type_k == 'string' then
						newname = name..format("[%q]",k) 
					elseif type_k == 'number' then
						newname = name..format("[%d]",k)
					else
						func = nil
					end
					if newname and not CheckTable(newname) then
						func = nil --can't open table so don't give function
					end
					catTable:AddLine('text', msg, 'func', func, 'arg1', self, 'arg2', newname, 'arg3', tstr)
				else
					(cats[type_v] or catOther):AddLine('text', msg, 'func', StaticPopup_Show, 'arg1', "SHOW_STRING_COPY", 'arg2', tostring(v))
				end
			end
         --ugly hack unceremoniously ripped from Nargiddley, Kemayo should be shot for this - so he delcares
         cat:AddCategory('text', L["Close"], 'justify', "CENTER", 'showWithoutChildren', true, 'func', ScheduleClose, 'arg1', tstr)
         tablet:SetHint(L["Click a table to view it. Ctrl-click to open in new window. Click other entry for access to copy."])
      end
   )
   tablet:Open(tstr)
end
